/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

/*
begin_html
<p>
 See also:
 <ul>
 <li><a href="max_ll_util.h.html">max_ll_util.h</a></li>
 <li><a href="ublas_matrixutil.h.html">ublas_matrixutil.h</a></li>
 <li><a href="callbk_reportstream.h.html">callbk_reportstream.h</a></li>
 </ul>
 <b>Up:</b> <a href="index.html">Index</a>
</p>
end_html
*/

#ifdef _WIN32
  #include <windows.h>
#else
  #include <stdlib.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <iostream>
#include <fstream>
using std::cout;
using std::cerr;
using std::endl;
#include <set>
using std::set;
#include <map>
using std::map;
#include <string.h>

#include "max_ll_util.h"
#include "ublas_matrixutil.h"
#include "callbk_reportstream.h"

#define MIN_TOTAL_RATE 1e-7
/*
#include <fstream>
std::ofstream logstream;

static std::ofstream& logg()
{
	if ( ! logstream.is_open() )
		logstream.open("C:\\qublog.txt");
	return logstream;
}
*/

/*
begin_html
  The Q matrix of rate constants (probability per second) is computed from
  intrinsic rate constants K0 and K1, and potentially Ligand- and Voltage-sensitive.
  \[Q_{a,b} = K0_{a,b} * Ligand_{a,b} * e^{K1_{a,b} * Voltage_{a,b}}\]
  \[Q_{a,a} = - \sum_i Q_{a,i}\]
end_html
*/
void BuildQ(int Ns, matrix<double>& Q, matrix<double>& K0, matrix<double>& K1, matrix<int>& Ligand, matrix<int>& Voltage, double *Constants)
{
  if ( Q.size1() != Ns || Q.size2() != Ns )
    Q.resize(Ns, Ns);
  for (int a=0; a<Ns; ++a) {
    double rowsum = 0.0;
    for (int b=0; b<Ns; ++b) {
      if (a != b) {
	double k = K1(a,b);
	if ( Voltage(a,b) )	  k *= Constants[ Voltage(a,b) ];
	k = exp(k);
	k *= K0(a,b);
	if ( Ligand(a,b) )
	  k *= Constants[ Ligand(a,b) ];
	if ( ! k )
	  k = MIN_TOTAL_RATE;
	Q(a,b) = k;
	rowsum += k;
      }
    }
    Q(a,a) = - rowsum;
  }
}

/*
begin_html
  "Qe" aka \({}^eQ\) is the "apparent" rate constant matrix, given that events
  with duration <= tdead are not recorded.  MIL computes the probability of
  staying in class <b>a</b> for duration <b>t</b> using the submatrix \({}^eQ_{aa}\)
  \[A(a, t) = e^{{}^eQ_{aa} (t-tdead)}\]
end_html
*/
void QtoQe(matrix<double>& Q, int *clazz, double tdead, matrix<double>& eQ)
{
  /*
    begin_html
   <p>
   eQ is partitioned into its Na x Nb submatrices, for classes a and b (likewise partition Q)
   each partition is computed separately.  Notation: going from class a to class b,
   z is "any class but a", and c is "any class but a or b".  Iz is the identity matrix with dimensions of Qzz.
   </p><p>
     \(a = b\):  \({}^eQ_{aa} = Q_{aa} - Q_{az} * (I_z - e^{Q_{zz} * tdead}) * Q_{za}\)
   </p><p>
   (accounts for going to another class for a short, then returning undetected)
   </p><p>
     \(a \ne b\):  \({}^eQ_{ab} = (Q_{ab} - (Q_{ac} * (I_c - e^{Q_{cc} * tdead}) * Q^{-1}_{cc} * Q_{cb}) * e^{Q_{bb} * tdead}\)
   </p><p>
   (accounts for going to class c for a short, on the way from a to b, then staying in b long enough to be detected)
   </p>
   end_html
  */

  int n = Q.size1();
  if ( eQ.size1() != n || eQ.size2() != n )
    eQ.resize(n, n);
  
  set<int> classet; // no duplicates, sorted increasing, for iteration
  for (int i=0; i<n; ++i)
    classet.insert(clazz[i]);
  for (set<int>::iterator cls_i = classet.begin(); cls_i != classet.end(); ++cls_i) {
    std::vector<int> states, outs;
    for (int i=0; i<n; ++i)
      if ( *cls_i == clazz[i] )
	states.push_back(i);
      else
	outs.push_back(i);
    int Na = states.size();
    int Nz = outs.size();
    int *sts = & (states[0]);
    int *ots = & (outs[0]);
    matrix<double> Qaa(Na, Na);        matrix_to_partition(Q, Na, Na, sts, sts, Qaa);
    matrix<double> Qaz(Na, Nz);        matrix_to_partition(Q, Na, Nz, sts, ots, Qaz);
    matrix<double> Qza(Nz, Na);        matrix_to_partition(Q, Nz, Na, ots, sts, Qza);
    matrix<double> Qzz(Nz, Nz);        matrix_to_partition(Q, Nz, Nz, ots, ots, Qzz);
    matrix<double> Iz(Nz, Nz);         fill_identity(Iz);
    matrix<double> Az(Nz, Nz);         expm(Qzz, tdead, Az);
    matrix<double> Qzi;                invm(Qzz, Qzi);
    matrix<double> missed_returning;   missed_returning.resize(Na, Na);
    noalias(missed_returning) = prod(matrix<double>(prod(matrix<double>(prod(Qaz, matrix<double>(Iz - Az))), Qzi)), Qza);
    matrix<double> eQaa(Na, Na);
    noalias(eQaa) = Qaa - missed_returning;
    matrix_from_partition(eQ, Na, Na, sts, sts, eQaa);
    
    for (set<int>::iterator cls_j = classet.begin(); cls_j != classet.end(); ++cls_j) {
      if ( *cls_j != *cls_i ) {
	std::vector<int> bees, cees;
	for (int i=0; i<n; ++i)
	  if ( *cls_j == clazz[i] )
	    bees.push_back(i);
	  else if ( *cls_i != clazz[i] )
	    cees.push_back(i);
	int Nb = bees.size();
	int Nc = cees.size();
	int *bs = Nb ? &(bees[0]) : NULL;
	int *cs = Nc ? &(cees[0]) : NULL;
	matrix<double> Qab(Na, Nb);    matrix_to_partition(Q, Na, Nb, sts, bs, Qab);
	matrix<double> Qbb(Nb, Nb);    matrix_to_partition(Q, Nb, Nb, bs, bs, Qbb);
	matrix<double> escaped_indirect(Na, Nb);
	if ( Nc ) {
	  matrix<double> Qac(Na, Nc);  matrix_to_partition(Q, Na, Nc, sts, cs, Qac);
	  matrix<double> Qcc(Nc, Nc);  matrix_to_partition(Q, Nc, Nc, cs, cs, Qcc);
	  matrix<double> Qcb(Nc, Nb);  matrix_to_partition(Q, Nc, Nb, cs, bs, Qcb);
	  matrix<double> Ic(Nc, Nc);   fill_identity(Ic);
	  matrix<double> Ac(Nc, Nc);   expm(Qcc, tdead, Ac);
	  matrix<double> Qci(Nc, Nc);  invm(Qcc, Qci);
	  noalias(escaped_indirect)  = prod( matrix<double>(prod( matrix<double>(prod(Qac, Ic - Ac)), Qci )), Qcb);
	}
	else {
	  // no third class? can't escape indirectly
	  fill_const(escaped_indirect, 0.0);
	}
	matrix<double> Ab(Nb, Nb);     expm(Qbb, tdead, Ab);
	matrix<double> eQab(Na, Nb);
	noalias(eQab) = prod(Qab - escaped_indirect, Ab);
	matrix_from_partition(eQ, Na, Nb, sts, bs, eQab);
      }
    }

  }
}

/*
begin_html
  <p>
  As discussed in (Milescu 2005), it can be preferable to start an experiment at equilibrium.
  Entry probabilities (P0) are then no longer constant, but a function of rate constants.
  We use the "direct" method given in Neher and Sakmann:
  </p><p>
  Defining \(S = [Q | 1]\) and \(u\) a row vector of ones,
  \[P_{eq} = u \cdot (S \cdot S^T)^{-1}\]
  </p>
end_html
*/
void QtoPe(matrix<double>& Q, matrix<double>& Pe)
{
  int N = Q.size1();
  if ( Pe.size1() != 1 || Pe.size2() != N )
    Pe.resize(1, N);
  matrix<double> S(N, N+1);
  matrix<double> U(1, N);
  for (int i=0; i<N; ++i) {
    for (int j=0; j<=N; ++j)
      S(i,j) = (j < N) ? Q(i,j) : 0.0;
    S(i,N) = 1.0;
    U(0,i) = 1.0;
  }
  matrix<double> SST(N, N);     noalias(SST) = prod(S, trans(S));
  matrix<double> SSTI(N, N);    invm(SST, SSTI);
  noalias(Pe) = prod(U, SSTI);
}




/*
 What are the optimization parameters?  (some of) the off-diagonal elements of K0 and K1.
 Actually, we use log(K0) to keep them positive.
 Breaking with past practice, we include all the unconnected zeros,
 and put all the K0s before all the K1s.  As we'll see, this is a column vector.
*/

inline int k_ix_of(int Ns, int a, int b)
{
  return (a * (Ns - 1)) + b - ((b > a) ? 1 : 0);
}

extern "C" MAXILL_API int __stdcall CountK(int Ns)
{
  return 2*(Ns*Ns - Ns); // the off-diagonal elements, twice
}

extern "C" MAXILL_API void __stdcall K01toK(int Ns, double **K0, double **K1, double *K)
{
  int Noffd = CountK(Ns) / 2;
  int i=0;
  for (int a=0; a<Ns; ++a)
    for (int b=0; b<Ns; ++b)
      if ( a != b ) {
	double x = K0[a][b];
	if ( x > 0.0 ) // tricky; 1->0 and 0->0; see KtoK01
	  x = log(x);
	else
	  x = 0.0;
	K[i] = x;
	K[Noffd+i] = K1[a][b];
	++i;
      }
}

// K alone isn't enough to reconstruct K0 and K1, since 0 could mean log(1), or else unconnected.
// We disambiguate by checking the original K0 for zeros, which mean unconnected.

#define MIN_LOG_K0 -30.0
#define MAX_LOG_K0 30.0

extern "C" MAXILL_API int __stdcall KtoK01(int Ns, double *K, double **K0, double **K1)
{
  int Noffd = CountK(Ns) / 2;
  int i = 0;
  for (int a=0; a<Ns; ++a)
    for (int b=0; b<Ns; ++b)
      if ( a != b ) {
	if ( K0[a][b] )
	  if ( (K[i] > MIN_LOG_K0) && (K[i] < MAX_LOG_K0) )
	    K0[a][b] = exp(K[i]);
	  else
	    return -(i+1);
	K1[a][b] = K[Noffd+i];
	++i;
      }
  return 0;
}

// But some of those K are zero and have to stay zero, and the user may have other constraints.
// We represent the constraints by the system of linear equations (11) and derive the transformations (12, 13).

// Some constraints are more a part of the representation than the model:
//   * where K0 == 0 there's no connection; fix k0 and k1
//   * where V == 0  fix k1

// Ncns_max = upper_bound(Ncns) = Nk [fixed] + 2*Ncns_user
// Ain[Ncns_max][Nk]
// Bin[Ncns_max]
// Ncns = 0 (or how many pre-existing constraints in Ain,Bin)

extern "C" MAXILL_API void __stdcall AddAutoConstraints(int Ns, double **K0, double **K1, int **L, int **V,
							int *Ncns, double **Ain, double *Bin)
// """Returns the constraints (Ain, Bin)   [11] which fix unconnected k0 and unsensitive k1."""
{
  int Nk = CountK(Ns);
  int Noffd = Nk/2;
  int i=0;
  for (int a=0; a<Ns; ++a)
    for (int b=0; b<Ns; ++b)
      if (a != b) {
        if ( ! K0[a][b] ) {
	  for (int j=0; j<Nk; ++j)
	    Ain[*Ncns][j] = 0.0;
	  Ain[*Ncns][i] = 1.0;
	  Bin[*Ncns] = 0.0;
	  V[a][b] = 0;
	  ++*Ncns;
	}
	if ( ! V[a][b] ) {
	  for (int j=0; j<Nk; ++j)
	    Ain[*Ncns][j] = 0.0;
	  Ain[*Ncns][Noffd+i] = 1.0;
	  Bin[*Ncns] = 0.0;
	  ++*Ncns;
	}
	++i;
      }
}

// The user might have some constraints too:

extern "C" MAXILL_API void __stdcall AddFixRateConstraint(int Ns, double **K0, double **K1, int **L, int **V,
							  int *Ncns, double **Ain, double *Bin,
							  int a, int b)
{
  int Nk = CountK(Ns);
  int k_ix = k_ix_of(Ns, a, b);
  
  for (int j=0; j<Nk; ++j)
    Ain[*Ncns][j] = 0.0;
  Ain[*Ncns][k_ix] = 1.0;
  Bin[*Ncns] = log(K0[a][b]);
  ++*Ncns;
}

extern "C" MAXILL_API void __stdcall AddFixExpConstraint(int Ns, double **K0, double **K1, int **L, int **V,
							 int *Ncns, double **Ain, double *Bin,
							 int a, int b)
{
  int Nk = CountK(Ns);
  int k_ix = k_ix_of(Ns, a, b);
  
  for (int j=0; j<Nk; ++j)
    Ain[*Ncns][j] = 0.0;
  Ain[*Ncns][k_ix + Nk/2] = 1.0;
  Bin[*Ncns] = K1[a][b];
  ++*Ncns;
}

// we optimize log(k0); this constraint actually means log(k0ab) - log(k0cd) = const

extern "C" MAXILL_API void __stdcall AddScaleRateConstraint(int Ns, double **K0, double **K1, int **L, int **V,
							    int *Ncns, double **Ain, double *Bin,
							    int a, int b, int c, int d)
{
  int Nk = CountK(Ns);
  int k_ix1 = k_ix_of(Ns, a, b);
  int k_ix2 = k_ix_of(Ns, c, d);
  
  for (int j=0; j<Nk; ++j)
    Ain[*Ncns][j] = 0.0;
  Ain[*Ncns][k_ix1] = 1.0;
  Ain[*Ncns][k_ix2] = - 1.0;
  Bin[*Ncns] = log(K0[a][b]) - log(K0[c][d]);
  ++*Ncns;
}

// ratio = (k1ab / k1cd)
// k1ab - ratio * k1cd = 0

extern "C" MAXILL_API void __stdcall AddScaleExpConstraint(int Ns, double **K0, double **K1, int **L, int **V,
							   int *Ncns, double **Ain, double *Bin,
							   int a, int b, int c, int d)
{
  int Nk = CountK(Ns);
  int k_ix1 = k_ix_of(Ns, a, b);
  int k_ix2 = k_ix_of(Ns, c, d);
  
  for (int j=0; j<Nk; ++j)
    Ain[*Ncns][j] = 0.0;
  Ain[*Ncns][k_ix1 + Nk/2] = 1.0;
  Ain[*Ncns][k_ix2 + Nk/2] = - K1[a][b] / K1[c][d];
  Bin[*Ncns] = 0.0;
  ++*Ncns;
}

extern "C" MAXILL_API void __stdcall AddGeneralizedConstraint(int Ns, int *Ncns, double **Ain, double *Bin,
															  double *Aentry, double Bentry)
{
	int Nk = CountK(Ns);
	for (int j=0; j<Nk; ++j)
		Ain[*Ncns][j] = Aentry[j];
	Bin[*Ncns] = Bentry;
	++*Ncns;
}

// nonzero if there's a problem, reported through report_cb
extern "C" MAXILL_API int __stdcall AddLoopBalConstraint(int Ns, double **K0, double **K1, int **L, int **V,
							  int *Ncns, double **Ain, double *Bin,
							  int Nloop, int *loopstates,
							  callbk_reportfun report_cb, void *report_data)
{
  callbk_reportstream output(report_cb, report_data);
  set<int> ligs, volts;
  map<int,int> ligsF, ligsB, voltsPlus, voltsMinus;
  
  int Nk = CountK(Ns);
  if ( loopstates[Nloop-1] == loopstates[0] )
    --Nloop; // don't repeat first as last
  
  for (int j=0; j<Nk; ++j)
    Ain[*Ncns][j] = 0.0;

  for (int i=0; i<Nloop; ++i) {
    int a = loopstates[i];
    int b = loopstates[(i+1)%Nloop];
    int k_ix1 = k_ix_of(Ns, a, b); 
    int k_ix2 = k_ix_of(Ns, b, a);
    
    Ain[*Ncns][k_ix1] = 1.0;
    Ain[*Ncns][k_ix2] = - 1.0;
    
    if ( L[a][b] ) {
      ligs.insert(L[a][b]);
      ++ligsF[L[a][b]];
    }
    if ( L[b][a] ) {
      ligs.insert(L[b][a]);
      ++ligsB[L[b][a]];
    }
    if ( V[a][b] ) {
      volts.insert(V[a][b]);
      if ( K1[a][b] < 0 )
	++voltsMinus[V[a][b]];
      else
	++voltsPlus[V[a][b]];
    }
    if ( V[b][a] ) {
      volts.insert(V[b][a]);
      if ( K1[b][a] < 0 )
	++voltsPlus[V[b][a]];
      else
	++voltsMinus[V[b][a]];
    }
  }
  Bin[*Ncns] = 0.0;
  
  for (set<int>::iterator ligi = ligs.begin(); ligi != ligs.end(); ++ligi) {
    if ( ligsF[*ligi] != ligsB[*ligi] ) {
      output << "In loop";
      for (int i=0; i<Nloop; ++i)
	output << " " << (loopstates[i]+1);
      output << ": Ligand (signal " << *ligi
	     << ") must influence the same number of rates forward and back." << endl;
      return -1;
    }
  }
  for (set<int>::iterator voli = volts.begin(); voli != volts.end(); ++voli) {
    if ( (! voltsPlus[*voli]) || (! voltsMinus[*voli]) ) {
      output << "In loop";
      for (int i=0; i<Nloop; ++i)
	output << " " << (loopstates[i]+1);
      output << ": Voltage (signal " << *voli
	     << ") must influence at least two rates,"
	     << " and at least one must have opposite sign or direction." << endl;
      return -2;
    }
  }
  
  ++*Ncns;
  
  if ( volts.size() ) {
    for (int j=0; j<Nk; ++j)
      Ain[*Ncns][j] = 0.0;
    for (int i=0; i<Nloop; ++i) {
      int a = loopstates[i];
      int b = loopstates[(i+1)%Nloop];
      int k_ix1 = k_ix_of(Ns, a, b); 
      int k_ix2 = k_ix_of(Ns, b, a);
      
      Ain[*Ncns][k_ix1 + Nk/2] = 1.0;
      Ain[*Ncns][k_ix2 + Nk/2] = - 1.0;
    }
    Bin[*Ncns] = 0.0;
    ++*Ncns;
  }
  return 0;
}

extern "C" MAXILL_API int __stdcall AddLoopImbalConstraint(int Ns, double **K0, double **K1, int **L, int **V,
							    int *Ncns, double **Ain, double *Bin,
							    int Nloop, int *loopstates,
							    callbk_reportfun report_cb, void *report_data)
{
  callbk_reportstream output(report_cb, report_data);
  set<int> ligs, volts;
  map<int,int> ligsF, ligsB, voltsPlus, voltsMinus;
  
  int Nk = CountK(Ns);
  if ( loopstates[Nloop-1] == loopstates[0] )
    --Nloop; // don't repeat first as last
  
  for (int j=0; j<Nk; ++j)
    Ain[*Ncns][j] = 0.0;

  double imbal = 0.0;
  for (int i=0; i<Nloop; ++i) {
    int a = loopstates[i];
    int b = loopstates[(i+1)%Nloop];
    int k_ix1 = k_ix_of(Ns, a, b); 
    int k_ix2 = k_ix_of(Ns, b, a);
    
    imbal += log(K0[a][b]) - log(K0[b][a]);
    
    Ain[*Ncns][k_ix1] = 1.0;
    Ain[*Ncns][k_ix2] = - 1.0;
    
    if ( L[a][b] ) {
      ligs.insert(L[a][b]);
      ++ligsF[L[a][b]];
    }
    if ( L[b][a] ) {
      ligs.insert(L[b][a]);
      ++ligsB[L[b][a]];
    }
    if ( V[a][b] ) {
      volts.insert(V[a][b]);
      if ( K1[a][b] < 0 )
	++voltsMinus[V[a][b]];
      else
	++voltsPlus[V[a][b]];
    }
    if ( V[b][a] ) {
      volts.insert(V[b][a]);
      if ( K1[b][a] < 0 )
	++voltsPlus[V[b][a]];
      else
	++voltsMinus[V[b][a]];
    }
  }
  Bin[*Ncns] = imbal;
  
  for (set<int>::iterator ligi = ligs.begin(); ligi != ligs.end(); ++ligi) {
    if ( ligsF[*ligi] != ligsB[*ligi] ) {
      output << "In loop";
      for (int i=0; i<Nloop; ++i)
	output << " " << (loopstates[i]+1);
      output << ": Ligand (signal " << *ligi
	     << ") must influence the same number of rates forward and back." << endl;
      return -1;
    }
  }
  for (set<int>::iterator voli = volts.begin(); voli != volts.end(); ++voli) {
    if ( (! voltsPlus[*voli]) || (! voltsMinus[*voli]) ) {
      output << "In loop";
      for (int i=0; i<Nloop; ++i)
	output << " " << (loopstates[i]+1);
      output << ": Voltage (signal " << *voli
	     << ") must influence at least two rates,"
	     << " and at least one must have opposite sign or direction." << endl;
      return -2;
    }
  }
  
  ++*Ncns;
  
  if ( volts.size() ) {
    for (int j=0; j<Nk; ++j)
      Ain[*Ncns][j] = 0.0;
    for (int i=0; i<Nloop; ++i) {
      int a = loopstates[i];
      int b = loopstates[(i+1)%Nloop];
      int k_ix1 = k_ix_of(Ns, a, b); 
      int k_ix2 = k_ix_of(Ns, b, a);
      
      Ain[*Ncns][k_ix1 + Nk/2] = 1.0;
      Ain[*Ncns][k_ix2 + Nk/2] = - 1.0;
    }
    Bin[*Ncns] = 0.0;
    ++*Ncns;
  }
  return 0;
}

/*
begin_html
<p>
 To derive the transformation we will use singular value decomposition (svd)
 which is really well described at http://www.uwlax.edu/faculty/will/svd/index.html
</p><p>
 And derive some useful extras:
 <ul>
   <li>\(W^{-1}\), where \(0^{-1} = 0\)</li>
   <li>\(A^{-1} = V W^{-1} U^T\)</li>
   <li>Nsv = number of nonzero singular values</li>
 </ul>
 </p>
end_html
*/

void DeleteConstraint(int Nk, int *Ncns, double **Ain, double *Bin, int ic)
{
  for (int jc=ic+1; jc<*Ncns; ++jc) {
    memcpy(Ain[jc-1], Ain[jc], Nk*sizeof(double));
    Bin[jc-1] = Bin[jc];
  }
  --*Ncns;
}

#define PRETTY_SMALL 1e-5

/*
begin_html
The <a href="http://www.uwlax.edu/faculty/will/svd/index.html">Singular Value Decomposition (SVD)</a>
technique in SetupLinearConstraints doesn't work well when
constraints duplicate each other; that is, none of the augmented vectors
\[c_i^T = \left [ A_{ij} \vdots B_i \right ]\]
may be a linear combination of the others.  To find and eliminate these constraints,
build C out of rows \(c_i\):  

\[C = \left [ A \vdots B \right ]^T\]

define the partitions:

\[C_{i..j} = \left [ {}^{c_{1i}}\ddots_{c_{N_k+1 j}} \right ]; C_{i..i} = c_i\]

and with i from (first+1) increasing, remove any column \(c_j\), where \(i \lt j\), if

\[C_{1..i} c'_j = c_j\]

can be solved for \(c'_j\).  This can be tested with SVD:

\[U W V^T = SVD(C_{1..i})\]
\[C^-1 \approx V W^{-1} U^T\]
<small>* all infinitesimals in U, W, V sent to 0<br>* if <math>W_{ii} = 0</math>, define <math>W_{ii}^{-1} = 0</math></small><br>

\[c'_j = C^{-1} c_j \approx V W^{-1} U^T c_j\]
\[c''_j = C_{1..i} c'_j\]

If \(c_j = c''_j\), then \(c_j\) is the linear combination \(c'_j\) of the previous constraints, so we remove it.
end_html
*/

extern "C" MAXILL_API void __stdcall ReduceConstraints(int Ns, int *Ncns, double **Ain, double *Bin)
{
  int Nk = CountK(Ns);
  fq::matrix<double> C(Nk+1, *Ncns);
  int ic = 0;
  int rank = 0;
  matrix<double> U, W, Vi, Csub;
  matrix<double> col(Nk+1, 1), re_col(Nk+1, 1);
  while ( ic < *Ncns ) {
    for (int i=0; i<Nk; ++i)
      C[i][ic] = Ain[ic][i];
    C[Nk][ic] = Bin[ic];
    if ( sv_decomp(C, Nk+1, ic+1, U, W, Vi) ) {
      int Nsv = 0;
      while ( (Nsv <= ic) && W(Nsv,Nsv) )
	++Nsv;
      if ( Nsv == rank ) { // check 1: added constraint must increase rank
	DeleteConstraint(Nk, Ncns, Ain, Bin, ic);
      } else {
	rank = Nsv;
/*
begin_html
As a double-check, the \(c_i\) are linearly independent if \(C_{1..N_c}\) has rank \(N_c\).
A constraint \(c_j\), where \(j = i+1\) can be removed if \(rank(C_{1..i}) = rank(C_{1..j})\).

Again with SVD: the singular values of \(C_{1..i}\) are the nonzero diagonal entries of W.
\(rank(C_{1..i})\) the number of singular values.
end_html
*/
	Csub.resize(Nk+1, ic+1);
	for (int i=0; i<=Nk; ++i)
	  Csub(i,ic) = C[i][ic];
	// prepare to invert Csub
	matrix<double> Wi(ic+1, ic+1);
	for (int i=0; i<=ic; ++i)
	  for (int j=0; j<=ic; ++j)
	    Wi(i,i) = ((i==j) && W(i,i)) ? (1.0/W(i,i)) : 0.0;
	matrix<double> Cinv(ic+1, Nk+1);
	noalias(Cinv) = prod(Vi, matrix<double>(prod(Wi, trans(U))));
	// Vi * Wi * U.T
	int jc = ic + 1;
	while ( jc < *Ncns ) {
	  // check 2: other constraints can't be linear combinations of the prior
	  for (int i=0; i<Nk; ++i)
	    col(i,0) = Ain[jc][i];
	  col(Nk,0) = Bin[jc];
	  matrix<double> image(ic+1, 1);
	  noalias(image) = prod(Cinv, col);
	  noalias(re_col) = prod(Csub, image);
	  double norm = 0.0;
	  for (int i=0; i<=Nk; ++i)
	    norm += DSQR(col(i,0) - re_col(i,0));
	  if ( norm < PRETTY_SMALL )
	    DeleteConstraint(Ns, Ncns, Ain, Bin, jc);
	  else
	    ++jc;
	}
	++ic;
      }
    } else {
      // svd failed so we'll just accept it i guess
      ++rank;
      Csub.resize(Nk+1, ic+1);
      for (int i=0; i<=Nk; ++i)
	Csub(i,ic) = C[i][ic];
      ++ic;
    }
  }
}


/*
begin_html
 Now that we have K0, K1, L, V and rate vector K,  
 we do some magic with the SVD to change your linear constraints
    \[A * K = B\]
 into a transformation between K and the column vector P of free parameters
    \[K = A*P + B\]
    \[P = Ainv * K\]

To solve the system for the subspace of constrained k, observe that any such k can be written

\[k = k_h + k_p\]

where \(k_h\) is in the null space of \(A_{in}\); i.e. \(A_{in} k_h = 0\), and \(k_p\) is a particular vector satisfying
\(A * K = B\).
end_html
*/

extern "C" MAXILL_API int __stdcall SetupLinearConstraints(int Nc, double **Ain, double *Bin,
							    int Nk, double *K, double *P_out, // Nk = Nc + Np
							    double **Acns, double *Bcns, double **Ainv,
								callbk_reportfun report_cb, void *report_data)
// Acns[Nk][Np]; Bcns[Nk]; Ainv[Np][Nk]
{
  callbk_reportstream output(report_cb, report_data);

  if ( Nc == 0 ) {
    for (int i=0; i<Nk; ++i) {
      Bcns[i] = 0.0;
      for (int j=0; j<Nk; ++j)
	Acns[i][j] = Ainv[i][j] = (i == j) ? 1.0 : 0.0;
      P_out[i] = K[i];
    }
    return 0;
  }

  matrix<double> U, W, Vi;
  if ( ! sv_decomp(Ain, Nc, Nk, U, W, Vi) ) {
    output << "Bad SVD." << endl;
    return -2;
  }
  int Nsv = 0;
  while ( (Nsv <= Nc) && W(Nsv,Nsv) )
    ++Nsv;
  if ( Nsv < Nc ) {
    output << "Bad SVD: Some constraints may be incompatible." << endl;
    return -2;
  }
  if ( Nsv != Nc ) {
    output << "Too many constraints.  "
	   << (Nc - Nsv) << " constraints may be linear combinations of the others." << endl;
    /*
begin_html
Consider \(rank(A_{in}^T) = rank(A_{in})\).  If \(rank(A_{in}) < rank(C)\),
there is a constraint that can be expressed as a linear combination of others,
except with a different right-hand-side (\(B_{in}\)) component, a contradiction. 
end_html
    */
    return -3;
  }
  for (int i=0; i<Nc; ++i)
    for (int j=Nc; j<Nk; ++j)
      if ( fabs(U(i,j)) > PRETTY_SMALL ) {
	output << "Bad SVD: 0 != U[c,non] == " << U(i,j) << endl;
	return -4;
      }

  /*
begin_html
==== The null space ====

Using SVD,

\[U W V^T = A_{in}\]
Now, assuming SVD sorts the singular values non-increasing, the columns of 

\[A = \left [ V_{R+1 \cdots N_k} \right ]\]
where R is the rank, or number of nonzero singular values
form a basis for the null space of <math>A_{in}</math>, so

\(k_h = A p = 0\) for all \(p \in \Re^{N_k-R}\)
end_html
  */
  int Np = Nk - Nc;
  
  for (int i=0; i<Nk; ++i)
    for (int j=Nc; j<Nk; ++j)
      Acns[i][j-Nc] = Vi(i,j);

  /*
begin_html
We find the particular nonhomogeneous solution <math>k_p</math> by inverting \(A_{in}\) using SVD pseudo-inversion:

\[A_{in} k_p = B_{in}\]
\[A_{in}^{-1} \approx V W^{-1} U^T\]
\[A_{in}^{-1} A_{in} k_p \approx V W^{-1} U^T B_{in}\]
\[B = k_p \approx V W^{-1} U^T B_{in}\]
end_html
  */
  matrix<double> Wi(Nk, Nk);
  Wi.clear();
  for (int i=0; i<Nk; ++i)
    Wi(i,i) = W(i,i) ? (1.0/W(i,i)) : 0.0;
  matrix<double> Ainvm = prod(Vi, matrix<double>(prod(Wi, trans(U))));
  matrix<double> Binm;
  matrix_from_array(Binm, Bin, Nc, 1);
  matrix<double> B = prod(Ainvm, Binm);
  for (int i=0; i<Nk; ++i)
    Bcns[i] = B(i,0);
  matrix<double> AcnsT(Np, Nk);
  for (int i=0; i<Nk; ++i)
    for (int j=0; j<Np; ++j)
      AcnsT(j,i) = Acns[i][j];
  matrix_to_arrays(AcnsT, (double**)Ainv, Np, Nk);
  matrix<double> K0;
  matrix_from_array(K0, K, Nk, 1);
  matrix<double> P = prod(AcnsT, K0);
  for (int i=0; i<Np; ++i)
    P_out[i] = P(i,0);

  return 0;
}

/*
begin_html
Optimizers poorly when components of p have different magnitude.
To scale them so they're all ones at the starting rate constants \(k0\), define the vector \(s\) of scaled parameters

\[p = A_s s + B_s\]

\[s = A^{-1}_s (p - B_s)\]

\[\frac{\partial p}{\partial s} = S\]

where \(A_s\) is the \(N_p x N_p\) diagonal matrix with entries

\[p0 = A^T k0\]

\[A_{s,ii} = \begin{cases} 
  1.0,  & \mbox{if }p0_i\mbox{ is 0} \\
  p0_i, & \mbox{otherwise} 
\end{cases}\]

and \(B_s\) is the column vector with entries

\[B_{s,i} = \begin{cases} 
  -1.0,  & \mbox{if }p0_i\mbox{ is 0} \\
   0.0,  & \mbox{otherwise} 
\end{cases}\]
end_html
*/

extern "C" MAXILL_API void __stdcall SetupStartAtOnes(int Np, double *pars, double **Ascal, double *Bscal, double **AscalInv)
{
  for (int i=0; i<Np; ++i)
    for (int j=0; j<Np; ++j)
      if (i != j) {
	Ascal[i][j] = 0.0;
	AscalInv[i][j] = 0.0;
      } else {
	if ( pars[i] ) {
	  Ascal[i][i] = pars[i];
	  Bscal[i]    = 0.0;
	  AscalInv[i][i] = 1.0 / pars[i];
	}
	else {
	  Ascal[i][i] = 1.0;
	  Bscal[i]    = -1.0;
	  AscalInv[i][i] = 1.0;
	}
      }
}

extern "C" MAXILL_API void __stdcall KtoPars(int Nk, int Np, double *K, double **A, double **Ainv, double *B, double *pars)
{
  matrix<double> AI(Np, Nk), KK(Nk, 1), PP(Np, 1);
  matrix_from_arrays(AI, Ainv, Nk, Np);
  matrix_from_array(KK, K, Nk, 1);
  noalias(PP) = prod(AI, KK);
  matrix_to_array(PP, pars, Np, 1);
}

extern "C" MAXILL_API void __stdcall ParsToK(int Np, int Nk, double *pars, double **A, double **Ainv, double *B, double *K)
{
  matrix<double> AA(Nk, Np), PP(Np, 1), KK(Nk, 1);
  matrix_from_arrays(AA, A, Nk, Np);
  matrix_from_array(PP, pars, Np, 1);
  noalias(KK) = prod(AA, PP);
  matrix_to_array(KK, K, Nk, 1);
  for (int i=0; i<Nk; ++i)
    K[i] += B[i];
}

/*
# To set up, you provide K0,K1,L,V and constraints, and we derive
# P and S and their transforms.  Then the optimizer adjusts S and we use it
# to build the Q matrices:
#   S --parsToK---> P --ParsToK---> K --KtoK01---> K0,K1 --BuildQ---> Q
*/
